from sklearn.model_selection import StratifiedKFold, KFold
from reval.relative_validation import RelativeValidation
from collections import namedtuple
from scipy import stats
import numpy as np
import math


class FindBestClustCV(RelativeValidation):
    """
    Child class of :class:`reval.relative_validation.RelativeValidation`.
    It performs k-fold cross validation on the training set to
    select the best number of clusters, i.e., the number that minimizes the
    normalized stability (i.e., average misclassification error/asymptotic misclassification rate).

    :param nfold: number of CV folds
    :type nfold: int
    :param nclust_range: list with minimum and maximum cluster numbers to look for
    :type nclust_range: list ([int, int])
    :param s: classification object inherited from :class:`reval.relative_validation.RelativeValidation`
    :type s: class
    :param c: clustering object inherited from :class:`reval.relative_validation.RelativeValidation`
    :type c: class
    :param nrand: number of random labeling iterations to compute asymptotic misclassification rate, inherited from
        :class:`reval.relative_validation.RelativeValidation` class
    :type nrand: int
    """

    def __init__(self, nfold, nclust_range, s, c, nrand):
        """
        Construct method
        """
        super().__init__(s, c, nrand)
        self.nfold = nfold
        self.nclust_range = nclust_range

    def best_nclust(self, data, strat_vect=None):
        """
        This method takes as input the training dataset and the
        stratification vector (if available) and performs a
        CV to select the best number of clusters that minimizes
        normalized stability.

        :param data: training dataset
        :type data: ndarray, (n_samples, n_features)
        :param strat_vect: vector for stratification, defaults to None
        :type strat_vect: ndarray, (n_samples,)
        :return: CV metrics for training and validation sets, best number of clusters,
            misclassification errors at each CV iteration
        :rtype: dictionary, int, dictionary
        """
        data_array = np.array(data)
        reval = RelativeValidation(self.class_method, self.clust_method, self.nrand)
        metrics = {'train': {}, 'val': {}}
        check_dist = {'train': {}, 'val': {}}
        for ncl in range(self.nclust_range[0], self.nclust_range[1]):
            if strat_vect is not None:
                kfold = StratifiedKFold(n_splits=self.nfold)
                fold_gen = kfold.split(data_array, strat_vect)
            else:
                kfold = KFold(n_splits=self.nfold)
                fold_gen = kfold.split(data_array)
            norm_stab_tr, norm_stab_val = [], []
            for tr_idx, val_idx in fold_gen:
                tr_set, val_set = data_array[tr_idx], data_array[val_idx]
                reval.clust_method.n_clusters = ncl
                miscl_tr, modelfit, tr_labels = reval.train(tr_set)
                miscl_val, val_labels = reval.test(val_set, modelfit)
                rndmisc_mean_val = reval.rndlabels_traineval(tr_set, val_set,
                                                             tr_labels,
                                                             val_labels)
                ms_val = miscl_val / rndmisc_mean_val
                norm_stab_tr.append(miscl_tr)
                norm_stab_val.append(ms_val)
                check_dist['train'].setdefault(ncl, list()).append(miscl_tr)
                check_dist['val'].setdefault(ncl, list()).append(ms_val)
            metrics['train'][ncl] = (np.mean(norm_stab_tr), _confint(norm_stab_tr))
            metrics['val'][ncl] = (np.mean(norm_stab_val), _confint(norm_stab_val))
        val_score = np.array([val[0] for val in metrics['val'].values()])
        bestscore = min(val_score)
        # select the cluster with the minimum misclassification error
        # and the maximum number of clusters
        bestncl = np.flatnonzero(val_score == bestscore)[-1] + self.nclust_range[0]
        return metrics, bestncl, check_dist

    def evaluate(self, data_tr, data_ts, nclust):
        """
        Method that applies the selected clustering algorithm with the best number of clusters
        to the test set. It returns clustering labels.

        :param data_tr: training dataset
        :type data_tr: ndarray, (n_samples, n_features)
        :param data_ts: test dataset
        :type data_ts: ndarray, (n_samples, n_features)
        :param nclust: best number of clusters
        :type nclust: int
        :return: labels and accuracy for both training and test sets
        :rtype: namedtuple, (train_cllab, train_acc, test_cllab, test_acc)
        """
        self.clust_method.n_clusters = nclust
        tr_misc, modelfit, labels_tr = super().train(data_tr)
        ts_misc, labels_ts = super().test(data_ts, modelfit)
        Eval = namedtuple('Eval',
                          ['train_cllab', 'train_acc', 'test_cllab', 'test_acc'])
        out = Eval(labels_tr, 1 - tr_misc, labels_ts, 1 - ts_misc)
        return out


def _confint(vect):
    """
    Private function to compute confidence interval.

    :param vect: performance scores
    :type vect: list
    :return: mean and error
    :rtype: tuple
    """
    error = stats.t.ppf(1 - (0.05 / 2), len(vect) - 1) * (np.std(vect) / math.sqrt(len(vect)))
    mean = np.mean(vect)
    return mean, error



import unittest
import numpy as np
from reval.best_nclust_cv import FindBestClustCV, _confint
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering
import math

# Modify to test other functions and parameters
RNDLABELS_ITER = 10
CLASSIFIER = KNeighborsClassifier(n_neighbors=5)
CLUSTERING = AgglomerativeClustering()
NCLUST_RANGE = [2, 4]
NFOLD = 2


class TestBestNclusterCV(unittest.TestCase):

    @classmethod
    def setUp(cls):
        cls.s = CLASSIFIER
        cls.c = CLUSTERING
        cls.nrand = RNDLABELS_ITER
        cls.nfold = NFOLD
        cls.nclust_range = NCLUST_RANGE
        cls.findbest = FindBestClustCV(cls.nfold, cls.nclust_range, cls.s, cls.c, cls.nrand)

    def test_best_nclust(self):
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        strat_vect = np.array([0, 1] * 20)
        metrics, best_nclust, _ = self.findbest.best_nclust(data,
                                                            strat_vect=strat_vect)
        m_tr = metrics['train'][best_nclust][0]
        m_val = metrics['val'][best_nclust][0]
        self.assertSequenceEqual([m_tr, m_val, best_nclust], [0, 0, 2])

    def test_evaluate(self):
        data_tr = np.array([[0] * 20,
                            [1] * 20] * 100)
        data_ts = np.array([[0] * 20,
                            [1] * 20] * 50)
        out = self.findbest.evaluate(data_tr, data_ts, 2)
        self.assertSequenceEqual([out.train_cllab.tolist(), out.train_acc,
                                  out.test_cllab.tolist(), out.test_acc], [[1, 0] * 100,
                                                                           1, [1, 0] * 50,
                                                                           1])

    def test_confint(self):
        dist = np.random.normal(0, 1, 1000000)
        m, error = _confint(dist)
        self.assertTrue(abs(round(m, 2)) <= 0.01 and
                        round(error * math.sqrt(len(dist)), 2) == 1.96)


if __name__ == 'main':
    unittest.main()



import pandas as pd
from bokeh.models import HoverTool, ColumnDataSource, CategoricalColorMapper
from bokeh.plotting import figure, show, output_notebook
from bokeh.io import export_png
from matplotlib import pyplot as plt


def plot_metrics(cv_score, title, figsize=(20, 10), save_fig=None):
    """
    Function that plots the average performance (i.e., normalized stability) over cross-validation
    for training and validation sets.

    :param cv_score: collection of cv scores as output by `reval.best_nclust_cv.FindBestCLustCV.best_nclust`
    :type cv_score: dictionary
    :param title: figure title
    :type title: str
    :param figsize: (width, height)
    :type figsize: tuple
    :param save_fig: file name for saving figure in png format, default None
    :type save_fig: str
    """
    fig, ax = plt.subplots(figsize=figsize)
    ax.plot(list(cv_score['train'].keys()),
            [me[0] for me in cv_score['train'].values()],
            linewidth=5,
            label='training set')
    ax.errorbar(list(cv_score['val'].keys()),
                [me[0] for me in cv_score['val'].values()],
                [me[1][1] for me in cv_score['val'].values()],
                linewidth=5,
                label='validation set')
    ax.legend(fontsize=18, loc=1)
    plt.xticks([lab for lab in cv_score['train'].keys()], fontsize=15)
    plt.yticks(fontsize=15)
    plt.xlabel('Number of clusters', fontsize=18)
    plt.ylabel('Normalized stability', fontsize=18)
    plt.title(title, fontsize=18)
    if save_fig is not None:
        plt.savefig(f'./{save_fig}', format='png')
    else:
        plt.show()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="reval",
    version="0.0.1",
    author="Isotta Landi",
    author_email="isotta.landi@iit.it",
    description="Relative clustering validation to select best number of clusters",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IIT-LAND/reval_clustering",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={"Documentation": "https://reval.readthedocs.io/en/latest/"},
    install_requires=["numpy",
                      "scipy",
                      "scikit-learn",
                      "umap-learn",
                      "matplotlib"],
    python_requires='>=3.6',
)
